/*
 * Open connection for network block device
 *
 * Copyright 1997,1998 Pavel Machek, distribute under GPL
 *  <pavel@atrey.karlin.mff.cuni.cz>
 *
 * Version 1.0 - 64bit issues should be fixed, now
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#define _USE_BSD
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>                  
#include <unistd.h>             /* get/setpgid stuff */

extern pid_t getpgid(pid_t pid);

#include <netinet/tcp.h>
#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */
#include <netdb.h>		/* hostent, gethostby*, getservby* */
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/time.h>
#include <sys/errno.h>

#ifdef USING_SSL
#include "sslincl.h"
#endif

#undef  MY_NAME
#define MY_NAME "nbd-client"
#ifndef __GNUC__
#error I need GCC to work
#endif

#include <linux/ioctl.h>
#define ISCLIENT
#include "cliserv.h"

/* lifted from fs.h */
#define BLKGETSIZE _IO(0x12,96) 

#define HELPME 1

#define MAXCONN 8
#define NBD_CLIENT_RENEGOTIATE 0x01
#define NBD_CLIENT_SIGNED      0x02
#define MAXDEV  8

struct nbd_client {
   char* hostname; /* PTB local host */
   int   port;     /* PTB port */
   int   saddr;    /* PTB local addr */
   int   daddr;    /* PTB dest addr */
   int   pid;      /* PTB daemon pid */
   int   slot;     /* PTB kernel slot */
   int* sig;       /* PTB imported device signature */
   int   i;        /* PTB client identifier */
   int   sock;     /* PTB open socket descriptor */
#ifdef USING_SSL
   SSL*  con;      /* AMARIN SSL open socket descriptor */
#endif
   int   dev;      /* PTB nbd device descriptor */
   int   flags;    /* PTB are we a client daemon? */
   int   ppid;     /* PTB parent pid */
   int   pgid;     /* PTB process group */
   u64 size;       /* PTB size of device in bytes */
   int   nmirror;
   int   mirror[MAXDEV];
# define NBD_CLIENT_WO 1
# define NBD_CLIENT_RO 2
# define NBD_CLIENT_RW 3
# define NBD_CLIENT_MR 4    /* PTB mirror */
# define NBD_CLIENT_DR 8    /* PTB direct */
   int   type;
};

static int   nport;              /* PTB total number of ports - shared */
static char *devicename[MAXDEV]; /* PTB whole device name - shared */
static int   ndevice;            /* number of devices (< 16) */
static int   blksize = 1024;     /* PTB device blocksize - shared, dflt 1024 */
static int   ppid;               /* PTB chief parent pid - shared */
static int   sig[NBD_SIGLEN/sizeof(int)];
                                 /* PTB whole device signature - shared */
#ifdef USING_SSL
static int   using_ssl = 0;      /* PTB modified later if we can and want */
#endif
static int   negotiate_timeout
                          = 60;  /* PTB timeout and restart in 60s */

// forward decls
int negotiate(struct nbd_client * self);
extern int errno;

#ifdef USING_SSL                 /* AMARIN shared variables for SSL */
extern int set_cert_stuff(SSL_CTX *ctx, char *cert_file, char *key_file);
static int verify_depth = 0;         
BIO* bio_err            = NULL;  
            /* AMARIN showing SSL errors - shared 
               BIO stands for blocking I/O */
static char* cert_file  = NULL;
static char* key_file   = NULL;
static char* CApath     = NULL;
static char* CAfile     = NULL;
static int verify       = SSL_VERIFY_NONE;
static SSL_METHOD* meth = NULL;
static SSL_CTX* ctx     = NULL;
static char* cipher     = NULL;
#endif

struct nbd_client clients[MAXCONN];
struct nbd_client manager = {
         0,    /* PTB hostname */
         0,    /* PTB local addr */
         0,    /* PTB dest addr */
         0,    /* PTB port */
         -1,   /* PTB daemon pid */
         -1,   /* PTB kernel slot */
         sig,  /* PTB slot signature */
         -1,   /* PTB client identifier */
         -1,   /* PTB open socket descriptor */
#ifdef USING_SSL
         0,    /* AMARIN SSL open socket descriptor */
#endif
         -1,   /* PTB nbd device descriptor */
         0,    /* PTB are we a client daemon? */
         0,    /* PTB parent pid */
         0,    /* PTB group pid */
         0,    /* PTB size of device in bytes */
         0,
         {0,0,0,0,},
    };
// list of pids for launcher and other fns with void args
int pids[MAXCONN];


// forward decl
void sigchldhandler();
void setsighandlers();

int microsleep(unsigned long n) {
    struct timeval delay = { n / 1000000, n % 1000000 };
    select(0,NULL,NULL,NULL,&delay);
    return 0;
}

int 
closenet (struct nbd_client * self)
{
    int err = 0;

#ifdef USING_SSL
    if (using_ssl) {
      if (self->con) err = SSL_shutdown(self->con);
    }
#endif

    err = close (self->sock);
    self->sock = -1;

#if defined(USING_SSL) && !defined(TRY_SSL_RECONNECT)
    if (using_ssl) {
      if (self->con) SSL_free(self->con);
      self->con = NULL;
    }
#endif

    MSG ("client (%d) closed socket %d\n", self->i, self->sock);
    return err;
}

int whoami()
  // PTB test list of pids to see if we can identify ourself
{
  int i = 0;
  int pid = -1;

  pid = getpid();
  if (pid == ppid)
     return -1;

  for (i = 0; i < nport ; i++) {
     if (!pids[i] || pids[i] == pid)
       return i;
  }

  return -2;
}

void
unplug_kernel ()
    // PTB unplug this client daemon from the kernel slot in which it is sitting
    // Done on exit so we have to figure out who we are
{
    int i =  whoami();

    if (i < 0) {
      return;  // parent or unknown
    } else {

      struct nbd_client * self = clients + i;

            // PTB - not if we didn't open a nbd device
      if (self->dev < 0)
        return;
            // PTB - not from an unsetup child daemon
      if (!(self->flags & NBD_CLIENT_RENEGOTIATE))
        return;
            // PTB it's OK from running daemons 
      MSG("unplug daemon (%d) from kernel slot %d and clear its socket\n",
          self->i, self->slot);
      ioctl (self->dev, NBD_CLEAR_SOCK);
    }
}


int
plug_kernel (struct nbd_client * self, int i)
    // PTB try and get slot i for sock in kernel
{
    if (self->dev < 0) 
      return self->slot = -1;
    if (self->sock < 0) 
      return self->slot = -1;
    self->slot = ioctl(self->dev, NBD_SET_SOCK, self->sock);
    return self->slot;
}

#ifdef USING_SSL
void 
closessl(struct nbd_client * self) {
   if (self->con) SSL_free(self->con);
   self->con = NULL;
}
#endif

void slavesighandler(int n) {

    int i = whoami();
    struct nbd_client * self = 0;

    if (i < 0)
      return;

    self = clients + i;

    PERR( "child sighandler received signal %d\n", n);

    switch (n) {
        case SIGTTIN: // PTB ignore
        case SIGTTOU:
        case SIGPIPE:
            MSG( "sighandler ignores exit %d in child client (%d) \n", n, self->slot);
            break;
        case SIGHUP:
            // PTB don't exit, just renegotiate
            MSG( "sighandler closes net for client (%d)\n", self->slot);
            closenet(self);
            if (self->dev >= 0) {
               MSG( "sighandler delivers resync from child\n");
               ioctl (self->dev, MY_NBD_SYNC, self->slot);
            }
            unplug_kernel();
            break;
        case SIGUSR1:
            MSG( "sighandler unplugs client (%d) from kernel\n", self->slot);
            unplug_kernel();
            if (self->dev >= 0) {
               MSG( "sighandler delivers soft reset from child\n");
               ioctl (self->dev, NBD_RESET);
            }
            break;
        case SIGUSR2:
            MSG( "sighandler unplugs client (%d) from kernel\n", self->slot);
            unplug_kernel();
            if (self->dev >= 0) {
               MSG( "sighandler delivers hard reset from child\n");
               ioctl (self->dev, NBD_HARD_RESET);
            }
            break;
        case SIGTERM:
        case SIGINT:
            MSG( "sighandler terminates child daemon\n");
            exit(0);
            break;
        case SIGCHLD:
        default:     // PTB ordinary exit will cause renegotiate
            MSG( "sighandler ignores exit %d in child client (%d) \n", n, self->slot);
            break;
    }
}

void propagate(int n) {
            int i;
            for (i = 0; i < nport; i++) {
              int pid = pids[i];
              if (pid > 0)
                kill(pid,n); // renegotiate
            }
}

void managersighandler(int n) {

    struct nbd_client * self = & manager;

    PERR( "manager sighandler received signal %d\n", n);

    switch (n) {

        case SIGHUP:
            MSG( "sighandler delivers device SYNC from manager\n");
            ioctl (self->dev, MY_NBD_SYNC, -1);
            propagate(n);
            break;
        case SIGTTIN: // PTB ignore
        case SIGTTOU:
        case SIGPIPE:
            MSG( "sighandler ignores exit %d in manager\n", n);
            break;
        case SIGCHLD:
            MSG( "sighandler relaunches child from manager \n");
            sigchldhandler(self);
            break;
        case SIGTERM:
        case SIGINT:
            propagate(n);
            MSG( "sighandler terminates manager safely\n");
            exit(0);
            break;
        case SIGUSR1:
            if (self->dev >= 0) {
               MSG( "sighandler delivers soft reset from manager\n");
               ioctl (self->dev, NBD_RESET);
            }
            propagate(n);
            break;
        case SIGUSR2:
            if (self->dev >= 0) {
               MSG( "sighandler delivers hard reset from manager\n");
               ioctl (self->dev, NBD_HARD_RESET);
            }
            propagate(n);
            break;
        default:     // PTB ignore
            PERR( "sighandler ignores exit %d in manager\n", n);
            break;
    }                     
}

#ifdef USING_SSL
   inline int ssl_check(SSL * con, int res) {

             switch (SSL_get_error(con,res)){ 
               case SSL_ERROR_WANT_WRITE:
               case SSL_ERROR_WANT_X509_LOOKUP:
               case SSL_ERROR_SYSCALL:
               case SSL_ERROR_ZERO_RETURN:
               case SSL_ERROR_SSL:
                 return -1;
              }
              return 0;
   }
#endif


#define WRITE 0x01
#define READ  0x00
#define CHCK  0x10
#define RCHCK CHCK|READ
#define WCHCK CHCK|WRITE

static fd_set fds, xfds;

/*
 *  Send or receive packet.
 */
int
xmit (int op, struct nbd_client* self, void *buf, int size)
{
    int result = 0;

    while (size > 0) {

         struct timeval net_timeout = { negotiate_timeout, 0 };
         struct timeval * timeout = & net_timeout;
           
        switch (op) {
	  case  WRITE: 
#ifdef USING_SSL
            if (using_ssl) {
              result = SSL_write(self->con, buf, size);
              if (ssl_check(self->con, result) < 0)
                return -1;
            } else
#endif
            {
               fd_set * wfds = & fds, * efds = & xfds;

               FD_ZERO(efds); FD_SET(self->sock,efds);
               FD_ZERO(wfds); FD_SET(self->sock,wfds);
             
               if (select(self->sock+1,NULL,wfds,efds,timeout) <= 0) 
                 return -1;
               result = send (self->sock, buf, size, 0);
            }

	    break;
	  default:
#ifdef USING_SSL
            if (using_ssl) {
              result = SSL_read(self->con, buf, size);
              if (ssl_check(self->con, result) < 0)
                return -1;
            } else
#endif
            {
               fd_set * rfds = & fds, * efds = & xfds;

               FD_ZERO(efds); FD_SET(self->sock,efds);
               FD_ZERO(rfds); FD_SET(self->sock,rfds);
             
               if (select(self->sock+1,rfds,NULL,efds,timeout) <= 0) 
                 return -1;
               result = recv (self->sock, buf, size, 0);
            }
	    break;
        }

	if (result <= 0)
	    break;

	size -= result;
	buf  += result;

    }
    return result;
}

long long llseek(int fd, long long offset,int whence) {

    long long remaining = offset;

    if (whence != SEEK_SET || offset < 0) // only SEEK_SET for now
       return -1ll;
    if (!(remaining >>31)) {
       int offset32 = offset;
       int res = lseek(fd,offset32,SEEK_SET);
       if (res < 0) return (long long)res;
       return offset;
    } else {
       int res = lseek(fd,(-1)>>1,SEEK_SET);
       if (res < 0) return (long long)res;
       remaining -= (-1)>>1;
    }
    while (remaining > 0) {
       if (!(remaining >>31)) {
          int offset32 = offset;
          int res = lseek(fd,offset32,SEEK_CUR);
          if (res < 0) return (long long)res;
          return offset;
       } else {
          int offset32 = (-1)>>1;
          int res = lseek(fd,offset32,SEEK_CUR);
          if (res < 0) return (long long)res;
          remaining -= offset32;
       }
    }
    return offset;
}

int
send_request (struct nbd_client* self, struct nbd_request *req)
{
    int result = 0;
    struct nbd_request request;

        /* PTB - nbd_request req has net order internally */

    request.magic = htonl(req->magic);
    request.type  = htonl(req->type);
    request.from  = htonll(req->from);
    request.len   = htonl(req->len);
    request.flags = htonl(req->flags);

    memcpy((char*) request.handle, (char*) req->handle, sizeof(request.handle)); /* PTB */

    switch (req->type & 0x01) {
      case READ:
        result = xmit (WRITE, self, &request, sizeof (struct nbd_request));
        break;

      case WRITE:
        result = xmit (WRITE, self, &request, sizeof (struct nbd_request));
        if (result <= 0) return result;

       /* PTB - this assumes that the data address is &req[1] */

        result = xmit (WRITE, self, req+1, req->len);
        if (self->nmirror > 0) {
           int i = 0;

              for (i = 1; i < self->nmirror; i ++) {
                 unsigned remaining = req->len;
                 unsigned long long offset = req->from;
                 long long res = llseek(self->mirror[i],offset,SEEK_SET);
                 if (res < 0) return res;
                 while (remaining > 0) {
                   result = write(self->mirror[i], req+1, remaining);
                   if (result <= 0)
                     return result;
                   remaining -= result;
                 }
              }
        }
        break;
    }

    return result;
}

int
read_stat (struct nbd_client* self, struct nbd_request *req, struct nbd_reply *rep)
{
    int result;
    char handle[NBD_HANDLE_LENGTH];
    int cmd;

    cmd = req->type & 0x03;

    memcpy ((char*)handle, (char*) req->handle, NBD_HANDLE_LENGTH); 

    if (cmd == READ && self->nmirror > 0) {
         unsigned remaining = req->len;
         unsigned long long offset = req->from;
         // should select. FIXME
         result = llseek(self->mirror[0],offset,SEEK_SET);
         rep->error = 1;
         rep->magic = NBD_REPLY_MAGIC;
         if (result <= 0)
             return result;
         while (remaining > 0) {
           result = read(self->mirror[0],rep+1,remaining);
           if (result <= 0)
             return result;
           remaining -= result;
         }
         rep->error = 0;
         memcpy ((char*)rep->handle, (char*)handle, NBD_HANDLE_LENGTH ); 
         return result;
    }

    // PTB - read the reply header
    result = xmit (READ, self, rep, sizeof (struct nbd_reply));

    // PTB - check result before magic!
    if (result <= 0)
      return result;

    // PTB convert interesting fields to host order
    rep->magic = ntohl(rep->magic);
    rep->error = ntohl(rep->error);

    // PTB - check the magic
    if (rep->magic != NBD_REPLY_MAGIC) {
       MSG("client (%d) reports bad magic 0x%x instead of 0x%x with "
            "error 0x%x handle %Lx flags %d cmd %d len %d\n", self->i,
            rep->magic, NBD_REPLY_MAGIC,
            rep->error, *(long long *) rep->handle, rep->flags, cmd, req->len);
       MSG("client (%d) reports addresses "
            "magic 0x%x error 0x%x handle 0x%x flags 0x%x\n", self->i,
            (unsigned)&rep->magic, (unsigned)&rep->error,
            (unsigned)rep->handle, (unsigned)&rep->flags);
       return -EINVAL;  // PTB should cause packet to be dropped 
    }
 
    if ((cmd & 0x01) == READ){
       // PTB - read some more bytes from stream to rest of buffer
       result = xmit (READ, self, rep+1, req->len);
    }

    memcpy ((char*)rep->handle, (char*)handle, NBD_HANDLE_LENGTH ); 
    return result;
}


int 
opennet (struct nbd_client * self)
{
    int sock;                 /* socket descriptor */
    struct sockaddr_in xaddrin;
    int xaddrinlen = sizeof (xaddrin);
    struct hostent *hostn;

    hostn = gethostbyname (self->hostname);
    if (!hostn)
	return -1;

    if ((sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
	return -2;

    self->sock = sock;
    PERR ("client (%d) opened socket %d\n", self->i, sock);

    xaddrin.sin_family = AF_INET;
    xaddrin.sin_port = htons (self->port);
    xaddrin.sin_addr.s_addr = *((int *) hostn->h_addr);
    if ((connect (sock, (struct sockaddr *) &xaddrin, xaddrinlen) < 0)) {
        PERR ("client (%d) socket connect failed %m\n", self->i);
	closenet (self);
	return -3;
    }

    setmysockopt (sock);

#ifdef USING_SSL
    if (using_ssl) {

      if (self->con)
#ifdef TRY_SSL_RECONNECT
        PERR("Reusing existing SSL connection...");  
#else
        closessl(self);
#endif
      if (!self->con)
        self->con = SSL_new(ctx);

      if (!SSL_set_fd(self->con,sock)) {
        PERR("SSL connection failed\n");  
        closenet(self);
        return -4;
      }

      if (SSL_connect(self->con) < 0) {
        PERR("SSL connection failed\n");  
        closenet(self);
        return -5;
      }

      PERR("SSL connection established\n");
    }
#endif
    return sock;
}

void usage(int err) {
        printf( 
            "Usage: host port .. [host port ..] .. [-b blksize] [-t timeout]"
#ifdef USING_SSL
            "[-cert certfile] [-key keyfile] "
            "[-CApath CApath] [-CAfile CAfile] "
            "[-verify depth] "
#endif
            "nbd_device [mirror_device ..]\n"
            "(error %d)\n", err
        );
}

int int_arg(int * res, char * a, int dflt) {

     char *err;
     *res = strtol (a, &err, 10);
     if (!*a || *err) {
        *res = dflt; 
        return -2;
     }
     return 0;
}


int 
cmdline (struct nbd_client * self, int argc, char *argv[])
{
    int i;
    char *hostname = 0;

    static struct { char *x;                       // name
                    char *t;                       // type
                    union {int * i; char **s;} a;  // target
                    union {int i; char *s;} d;     // default
                    char *m;                       // message
                  } options[] =
    {
     {"b", "i", {&blksize}, {1024}, MY_NAME ": " "client says blksize is %d\n"},
     {"t", "i", {&negotiate_timeout}, {60}, MY_NAME ": " "client negotiation timeout is %d\n"},
#ifdef USING_SSL
     {"cert", "s+", {&cert_file}, {NULL}, MY_NAME ": " "client cert file is %s\n"}, 
     {"cipher", "s+", {&cipher}, {NULL}, MY_NAME ": " "client cipher: %s\n"},
     {"keyfile", "s+", {&key_file}, {NULL}, MY_NAME ": " "key cert file is %s\n"},
     {"CApath", "s+", {&CApath}, {NULL}, MY_NAME ": " "client CApath is %s\n"},
     {"CAfile", "s+", {&CAfile}, {NULL}, MY_NAME ": " "client CAfile is %s\n"},
     {"verify", "i+", {&verify_depth}, {0}, MY_NAME ": " "client verify_depth is %d\n"},
#endif
     {NULL,NULL,{NULL},{0},NULL},
    };

    if (argc < 4) {
        return 10;
    }

    for (i = 1 ; i < argc; i++) {

        if (hostname) {
        /* look for a port */
          int err = 0;
          unsigned long port;
          err = int_arg ((int*)&port, argv[i], -1);
          if (err < 0)  {
	    /* it's a new hostname or a device */
	    if (i < argc - 1 && *argv[i] != '-' ) {
	      /* hostname */
              hostname = argv[i];
              continue;
            }
            /* device or option */
            if (*argv[i] == '-' ) { 
              /* option */
              int j; int success = 0;
              for (j = 0; options[j].x ; j++) {
                 if (argv[i][1] != options[j].x[0])
                    continue;
                 if (i+1 >= argc) 
                    return 1;
                 if (strcmp(argv[i]+1,options[j].x))
                    continue;
                 switch (options[j].t[0]) {
                   int err;
                   case 'i': 
                      err = int_arg(options[j].a.i,argv[++i],options[j].d.i);
                      if (err < 0)
                        return 2;
#ifdef USING_SSL
                      if (options[j].t[1]) /* == '+' */
                         using_ssl = 1;
#endif
                      syslog(LOG_DAEMON|LOG_INFO|LOG_CONS|LOG_PERROR,
                       options[j].m, * options[j].a.i);
                      success = 1;
                      break;
                    case 's':
                      *options[j].a.s = argv[++i];
#ifdef USING_SSL
                      if (options[j].t[1])
                         using_ssl = 1;
#endif
                      syslog(LOG_DAEMON|LOG_INFO|LOG_CONS|LOG_PERROR,
                       options[j].m, *options[j].a.s);
                      success = 1;
                      break;
                 } /* end sw */
                 if (success)
                   break;
              } /* end for j */
              if (!success)
	        return 9;
              /* next option i */
              continue;
	    }
	    /* device */
            devicename[ndevice++] = argv[i];
            MSG("client device name is %s\n",devicename[ndevice-1]);
            continue;
          } else {
	    /* it is a port */
            clients[nport].port = port;
            clients[nport].hostname = hostname;
            clients[nport].type = NBD_CLIENT_RW ;
            MSG("client says host name %d is %s\n",nport,hostname);
	    MSG("client says port %d is %ld\n",nport,port);
            /* end of one client option set */
            clients[nport].i    = nport++;
	  }
	  continue;
        } 
        /* look for a hostname */
        hostname = argv[i];
    }

    return 0;

}


int getproto(struct nbd_client * self) {
              int err = 0;
#ifdef MY_NBD_REG_BUF
              int old = 0;
              int dev = self->dev;
              if (err = ioctl(dev, MY_NBD_REG_BUF, 0), err < 0) 
                  old = 1;
#else
              int old = 1;
#endif
              return old;
}

int server_check(struct nbd_client* self, struct nbd_request * req, struct nbd_reply * rep) {
    // PTB make dummy request of type READ to send out
    int err = 0;
    int i = NBD_HANDLE_LENGTH;
    static u64 count;

    req->magic = NBD_REQUEST_MAGIC;
    // PTB indicate to server it's a keep alive
    req->type  = RCHCK;
    // PTB zero handle. memset(req->handle,0,sizeof(*req->handle)*NBD_HANDLE_LENGTH)
    while (--i >= 0) 
       req->handle[i] = 0;
    // PTB increment offset by one each time
    req->from  = (count++ << 9) % self->size;
    req->len   = 0;
    err = send_request(self, req);
    if (err < 0)
       return err;
    err = read_stat(self, req, rep);
    return err;
}

int newproto(struct nbd_client * self, char* ioctl_buf) {
                  int err = 0;

                  struct nbd_request *request;
                  struct nbd_reply *reply;
		  request = (struct nbd_request*)ioctl_buf;
		  reply   = (struct nbd_reply*) ioctl_buf;

                  get_req:
	          err = ioctl(self->dev, MY_NBD_GET_REQ);   // PTB  

 	          if (err < 0) {
                      static int count = 0;
                      err = errno;
                      if (err != ETIME)
		        PERR ("kernel GET_REQ call returned: %m (%d)\n", err);
                      // PTB kernel thought we already had one or inactivity timeout
                      // MSG ("Server_check (%d) made and sent\n", self->i);
                      err = server_check(self, request, reply);
                      if (err < 0)  {
                          PERR ("Server_check (%d) failed %m on %d so clear socket\n",self->i,self->sock);
                          ioctl (self->dev, NBD_CLEAR_SOCK); 
                          return err = -1; // PTB renegotiate
                      }
                      // MSG ("dummy CHK_REQ (%d) received and discarded OK\n", self->i);
                      if (count++ < 3)
                      // PTB three times in four, repeat quickly
                         goto get_req;
                      // PTB one time in four, pause
                      count = 0;
                      microsleep(1000000);
                      // PTB and rollback nonexistent kernel request and retry for new
                      ioctl (self->dev, MY_NBD_CLR_REQ); 
                      return err = 0; // PTB could be goto get_req
	          }

                  send:
		  err = send_request(self, request);
                  
                  // PTB need to spot socket error and go to renegogiate

		  if (err < 0) {
                        static int count = 0;
                        switch(errno) {
                          case SIGPIPE:
                          case -SIGPIPE:
                          default:
                            if (count++ < 3) {
                              PERR ("Send_request (%d) failed %m on %d so retry %d\n",self->i,self->sock,count);
                              microsleep(1000000);
                              goto send;
                            }
                            count = 0;
                        }
                        PERR ("Send_request (%d) failed %m on %d so clear socket\n",self->i,self->sock);
                        ioctl (self->dev, NBD_CLEAR_SOCK); 
                        return err = -1;             // PTB renegogiate
 	          }

                  err = read_stat(self, request, reply);

		  if (err < 0) {
		        switch (err) {
                        case -EINVAL:
                          PERR ("read_stat errored on magic: %m\n");
                          // PTB rollback kernel request, discard our own
                          ioctl (self->dev, MY_NBD_CLR_REQ); 
                          // PTB there are bytes left on the input stream
                          return err = 0;
		        default:
                          PERR ("Read_stat (%d) failed %m on %d so clear socket\n",
                             self->i, self->sock);
                          ioctl (self->dev, NBD_CLEAR_SOCK); 
                          return err = -1;             // PTB renegogiate
                        }
 	          }

		  err = ioctl(self->dev, MY_NBD_ACK);   // PTB 

		  if (err < 0) {
                        // PTB kernel thinks we're crazy
                        PERR ("Kernel ACK call returned: %m\n");
                        // PTB rollback kernel request, discard our own
                        ioctl (self->dev, MY_NBD_CLR_REQ); 
                        return err = 0;
 	          }
                  return err;
}

void setsighandlers(void sighandler(int)) {
     // PTB handle all signals in one sighandler
     int k = 0;
     for (k = 1; k < 30; k++) {
#if defined(__GLIBC__) && __GLIBC__ >= 2
            struct sigaction sa = {{sighandler}, {{0}}, SA_RESTART, NULL};
#else
            struct sigaction sa = {{sighandler}, 0, SA_RESTART, NULL};
#endif
            sigfillset(& sa.sa_mask);
            sigdelset(& sa.sa_mask, SIGTERM); // PTB let TERM kill handler
            sigdelset(& sa.sa_mask, SIGINT);  // PTB let INT kill handler
            sigaction(k, & sa, NULL);
     }
     //signal(SIGTERM,  SIG_DFL);            // PTB handle TERM naturally
     //signal(SIGINT,   SIG_DFL);            // PTB handle INT naturally
}

void resetsighandlers() {
     // PTB restore defaults
     int k = 0;
     for (k = 1; k < 30; k++) {
#if defined(__GLIBC__) && __GLIBC__ >= 2
            struct sigaction sa = {{SIG_DFL} , {{0}}, SA_RESTART, NULL};
#else
            struct sigaction sa = {{SIG_DFL} , 0, SA_RESTART, NULL};
#endif
            sigfillset(& sa.sa_mask);
            sigaction(k, & sa, NULL);
     }
}

int getslot(struct nbd_client * self) {

          int err = 0;

          int k = self->i;

          MSG( "client (%d) hands down socket %d to device %s in slot %d\n",
              self->i,self->sock,devicename[0], k);
          err = plug_kernel(self, k);
          if (err < 0) {
	    PERR ("client could not get any vacant socket slot from kernel");
            PERR ("Ioctl/2 failed: %m\n");
            return self->slot = -1;
          }

          return self->slot = self->i;
}

int getblksize(struct nbd_client * self) {
       /* PTB if we don't yet know kernel devices blksize, get it */
       int blksize;
       ioctl(self->dev, NBD_GET_BLKSIZE, &blksize);
       return blksize;
}
int setblksize(struct nbd_client * self, int blksize) {
       int err = 0;
       /* PTB if kernel device doesn't yet know its blksize, set it */
       if( ioctl(self->dev, NBD_GET_BLKSIZE, &err) < 0 || err != blksize) {
             if (ioctl (self->dev, NBD_SET_BLKSIZE, blksize) < 0)
	       return -1;
       }
       return blksize;
}


int setsize(struct nbd_client * self, u64 size) {
       int err = 0;
       /* PTB if kernel device doesn't yet know its size, set it */
       if( ioctl(self->dev, BLKGETSIZE, &err) < 0 || err + 1 < 0) {
         unsigned sectors = size >> 9;
         MSG("client (%d) device not yet set to size\n", self->i);
         if (err = ioctl (self->dev, NBD_SET_SECTORS, sectors), err < 0)  {
           PERR("client (%d) failed to set size %Ld\n", self->i, size);
	   return -1;
         }
         MSG("client (%d) set device size %Ld\n", self->i, size);
         return 0;
       } 
       size = err;
       size <<= 9;
       MSG("client (%d) kernel says size is %Ld\n", self->i, size);
       return 0;
}

  inline void bounce(unsigned long sig) {
     static jmp_buf * trampoline = 0;
     if (sig < 4096) {
       jmp_buf * j = trampoline;
       trampoline = 0;
       if (j)
         longjmp(*j, sig);
     } else {
       trampoline = (jmp_buf *)sig;
     }
  }

  inline void set_alarm(unsigned long timeout) {
         static void (*sig)(int);
         alarm(timeout);
         if (timeout > 0) {
           // PTB sigalarm will make us fail to the trampoline return
           sig = signal(SIGALRM, (void(*)(int))bounce);
         } else {
           // PTB sigalarm is ignored
           signal(SIGALRM, sig);  // PTB restore old handler
         }
   }
 
int mainloop(struct nbd_client * self) {

              // PTB this is the kernel/net protocol loop. Repeat until done.

              int old = 0;
              int err = 0;
              char ioctlbuf[NBD_MAX_SECTORS*512 + 1000] ; /* AMARIN */
          
              old = getproto(self);
              MSG("Client (%d) chooses %s protocol\n",
                      self->i, old?"old":"new");

              switch (old) {

                case 0:  /* PTB new style user space net connex */
                if (err = ioctl(self->dev, MY_NBD_REG_BUF, ioctlbuf), err < 0)
                    break;
	        MSG ("begin main loop (%d)\n", self->i);
	        do {
                  // PTB new 60s timer code
                  static jmp_buf my_jmp;

                  bounce((unsigned long)&my_jmp);
                  if (setjmp(my_jmp)) {
                    // PTB timed out - bounced here on SIGALRM
                    PERR("client (%d) times out waiting 60s in mainloop\n",
                      self->i);
                    err = -1;
                    break;  // PTB return -1
                  }
                  // MSG("Client set 60s timer for mainloop\n");
                  set_alarm(60);

                  err = newproto(self, (char*)ioctlbuf);

                } while (!err);

                // MSG("Client unset timer for mainloop\n");
                set_alarm(0);
                break;

                case 1:  /* PTB old style kernel space net connex */
	        MSG ("begin main loop (%d)\n", self->i);
	        do {
                  err = ioctl (self->dev, NBD_DO_IT); 
	        } while (!err);
                break;
              }
              return err;
}
	
int launch(struct nbd_client * self, int clr_sock) {

    // async launch of a process to guard the client struct
    // At the time we run we have already forked off from the main
    // process so we don't share new opens with its other children.

    int err = 0;
    int pid = -1;
    char buffer[1024];

    if (self->dev < 0)          // PTB paranoia: did we get a whole device?
        return self->dev;
                                // PTB close whole device
    close(self->dev);        
                                // PTB open subdevice

    sprintf(buffer,"%s%d",devicename[0],self->i+1);
    self->dev = open (buffer, O_RDWR);
    if (self->dev < 0) {
        PERR("childminder failed to open subdevice %s for daemon (%d)\n",
            buffer,self->i);
        return self->dev;
    }
    MSG("childminder opens subdevice %s for daemon (%d)\n",buffer,self->i);

    self->flags &= ~NBD_CLIENT_RENEGOTIATE;

    if (clr_sock && self->slot >= 0) {
       MSG("childminder replacing daemon (%d) %d clears socket\n",
         self->i,self->pid);
       ioctl (self->dev, NBD_CLEAR_SOCK);
    }
    self->pid  = -1;

    pid = fork();
    if (pid < 0) {
        PERR ("client manager daemon %d launch failed on port %s:%d\n",
                       self->i, self->hostname, self->port);
        return self->pid = pid;
    }
    if (pid > 0) {              /* parent returns with child pid */
        MSG ("client manager launched daemon %d (%d) on port %s:%d\n",
                       self->i, pid, self->hostname, self->port);
        return self->pid = pid;
    }

    // child

    resetsighandlers();
    self->pid = getpid();
    pids[self->i] = self->pid; // so whoami() works
    self->sock = -1;

    // PTB open mirror devices.

    self->nmirror = 0;
    while (self->nmirror < ndevice - 1) {
       self->mirror[self->nmirror] = open(devicename[self->nmirror + 1],O_RDWR);
       self->nmirror++;
    }

    // PTB go somewhere safe

    chdir ("/");

    do {
       // PTB repeatedly try to set up initial connection 

      do {
        err = negotiate(self);
      } while (err < 0 && (microsleep(10000000),1)); // PTB delay on fail

       // do protocol
      err = mainloop(self);

      // PTB see if we want to renegotiate when the protocol fails
      switch (err) {
	  case -1:
              self->flags |= NBD_CLIENT_RENEGOTIATE;
	      err = 0;
	      break;
          default:
              self->flags &= ~NBD_CLIENT_RENEGOTIATE;
	      break;
      }
    } while ( (self->flags & NBD_CLIENT_RENEGOTIATE)
            && (microsleep(10000000),1)); // PTB delay on fail

     // PTB child death. try and close down socket only

    if (self->sock >= 0) {
      	 PERR("child daemon closes open socket on death\n");
         closenet(self);
         // PTB CLEAR_SOCK done at exit
    }
    microsleep(5000000); // PTB delay to avoid fast SIGCHLDs in sequence
    return 0;
}


   
int negotiate(struct nbd_client * self) {

            int err = 0;
            u64 size64;
            u64 magic;
            char buf[256] = "\0\0\0\0\0\0\0\0\0";
            unsigned long blksize_host = blksize;
            jmp_buf my_jmp;
            unsigned long timeout = negotiate_timeout;

              bounce((unsigned long)&my_jmp);
              if (setjmp(my_jmp)) {  
                // PTB timed out - bounced here on SIGALRM
                err = -15;
                goto fail;
              }
    
              set_alarm(timeout);

              // PTB join inits pgroup to spare parent a SIGCHLD if we die early

              setpgid(0,getpgid(1));

              // PTB we no longer handle minor errors in the first
              //     negotiate seq internally

	      err = 0;
              resetsighandlers();          // PTB signals -> defaults

              if (self->sock >= 0) {
        	 MSG("launch closes already open client socket %d\n",
                     self->sock);
                 closenet(self);
	      }

	      if (err = opennet (self), err < 0) {
                  if (!(self->flags & NBD_CLIENT_RENEGOTIATE)) 
		      switch (err) {
			case -1:
			    PERR ("Gethostname failed: %m\n");
                            err = -1; goto fail;
			case -2:
			    PERR ("Socket failed: %m\n");
                            err = -1; goto fail;
			case -3:
			    PERR ("Connect failed: %m\n");
                            err = -1; goto fail;
			case -4: case -5:
			    PERR ("SSL connection failed: %m\n");
                            err = -1; goto fail;
		      }
                  // PTB error is sustainable and could try renegotiate
                  err = -1; goto fail;
	      }

              MSG("client (%d) opened socket %d\n",self->i,self->sock);

	      MSG ("client (%d) negotiation begins\n", self->i);

              if (xmit(READ, self, buf, 8) < 0) {
		      PERR ("Failed/1: %m\n");
                      err = -1; goto fail;
	      }
	      if (!HELPME && strcmp (buf, INIT_PASSWD)) {
		      PERR ("INIT_PASSWD bad\n");
                      err = -1; goto fail;
	      }

              MSG("client (%d) read passwd ok\n",self->i);

	      /* PTB we most often die here on retrying, if we die */

              if (xmit (READ, self, &magic, sizeof (magic)) < 0) {
		      PERR ("Failed/2: %m\n");
                      err = -1; goto fail;
	      }
	      magic = ntohll (magic);
	      if (magic != cliserv_magic) {
		      PERR ("Not enough cliserv_magic\n");
                      err = -1; goto fail;
	      }

              MSG("client (%d) got cliserv magic ok\n",self->i);

              if (xmit (READ, self, &size64, sizeof (size64)) < 0) {
		      PERR ("Failed/3: %m\n");
                      err = -1; goto fail;
	      }
	      size64 = ntohll (size64);
#if 0
              if (size64 != self->size) {
		      MSG ("client says size %Ld != size %Ld\n",
                              self->size, size64);
                      err = -1; goto fail;
              }
#endif

	      self->size = size64;

	      MSG ("client (%d) got size = %Lu", self->i, self->size);

	      if (1) {
		  int rebuf[NBD_SIGLEN/sizeof(int)] = {0,0,0,0,0,0,0,0,};
                  if (xmit (READ, self, (char*)rebuf, NBD_SIGLEN) < 0) {
			  PERR ("Failed/4: %m\n");
                          err = -1; goto fail;
		  }

		  MSG ("client (%d) got signature 0x%x.. in %lx,"
                       " had 0x%x.. in %lx\n", self->i,
                                    * rebuf, (unsigned long)rebuf,
                                    * self->sig, (unsigned long) self->sig);

                  if (ioctl (self->dev, MY_NBD_SET_SIG, rebuf) < 0) {
	                  PERR ("Failed sig check\n");
                          err = -1; goto fail;
                  }

		  MSG ("client (%d) passed or set sigcheck\n", self->i);

                  if (!(self->flags & NBD_CLIENT_SIGNED)) {
		      MSG ("client (%d) wants to write 128 bytes "
                           "from %lx to %lx\n", self->i,
                             (unsigned long)rebuf, (unsigned long) self->sig);

		      memcpy ((char*) self->sig, (char*)rebuf, NBD_SIGLEN);

                      self->flags |= NBD_CLIENT_SIGNED;
		      MSG ("client (%d) set internal sig copy\n", self->i);
                  }
	      }

              if (xmit (WRITE, self, &blksize, sizeof(blksize)) < 0) {
			  PERR ("Failed/5 on descriptor %d: %m\n", self->sock);
                          err = -1; goto fail;
              }

              MSG ("client (%d) suggested blksize %d\n", self->i, blksize);

              if (xmit (READ, self, &blksize_host, sizeof(blksize_host)) < 0) {
			  PERR ("Failed/6: %m\n");
                          err = -1; goto fail;
              }

              // PTB we choose the MAX of the two offered blksizes

              MSG ("client (%d) received suggestion blksize %ld\n",
                   self->i, blksize_host);

              if (blksize_host > blksize)
                  blksize = blksize_host;

              MSG ("client (%d) negotiated blksize %d\n", self->i, blksize);

              if (setsize(self, self->size) < 0) {
		  PERR ("Ioctl/1 failed: %m\n");
                  err = -1; goto fail;
              }
              
              MSG ("client (%d) set size %Lu\n", self->i, self->size);

              if (blksize) {
                if (setblksize(self, blksize) < 0) {
		    PERR ("Ioctl/1.1 failed: %m\n");
                    err = -1; goto fail;
                }

                MSG ("client (%d) set blksize %u\n", self->i, blksize);

              } else {

                /* PTB ask kernel if we don't know */

                blksize = getblksize(self);
                if (blksize < 0) {
		    PERR ("Ioctl/1.2 failed: %m\n");
                    err = -1; goto fail;
                }

                MSG ("client (%d) discovered blksize %u\n", self->i, blksize);

              }

              set_alarm(0);                // PTB all over. Cancel alarm.

              // PTB tell the kernel we're here
	       
              if (getslot(self) < 0) {
                    PERR ("Ioctl/2 failed: %m\n");
                    err = -1; goto fail;
              }

              MSG(
                "client (%d) set device %s to use our socket %d in slot %d\n",
                self->i,devicename[0],self->sock,self->slot);

              /* PTB - try and set termination action to clear socket */

              atexit(unplug_kernel);

	      /* Go daemon */

              self->flags |= NBD_CLIENT_RENEGOTIATE;
              setpgid(0,self->pgid);       // PTB rejoin parents process group
              setsighandlers(slavesighandler); // PTB handle CHLD USR1 HUP TERM

              return 0;

fail:
              set_alarm(0);                // PTB all over. Cancel alarm.
              resetsighandlers();          // PTB signals -> defaults

       	      PERR("negotiate bails out on client socket %d\n", self->sock);

              closenet(self);
              return err;

}


int restart(struct nbd_client * self, int died) {
  
    // PTB check and restart a known pid. (or all if pid = -1)

    int i = 0;
    int count = 0;

    for ( i = 0; i < nport; i++)  {

         int pid = pids[i];

         MSG("main childminder checking pid %d\n", pid);

         if ( pid < 0 || kill(pid,0) < 0 || died == pid )  {

             pids[i]  = -1;

             pid = launch(& clients[i],1);
             if (pid > 0) {
               pids[i] = pid;
               count ++;
             }
         }
    }
    return count;
}

void sigchldhandler(struct nbd_client * self) {

    // PTB reap and restart dead children.

    int died = 0;
    int count = 0;
    
    if (self->dev <= 0)
        return;

    while ( died = wait4(-1,0,WNOHANG,0), died > 0 )
      count += restart(self, died) ;

    if (count <= 0) 
      // PTB a mysterious unknown child died so check all
      count += restart(self, -1) ;

}

void init(struct nbd_client * self, int i, int dev, int * sig, int ppid, int pgid) {
        self->i    = i;
        self->dev  = dev;
        self->sig  = sig;
        self->ppid = ppid;
        self->pgid = pgid;
#ifdef USING_SSL
        self->con  = NULL;
#endif
        self->flags= 0;
}

#ifdef USING_SSL
int
init_ssl(struct nbd_client * self) {
      /* AMARIN initializing SSL stuff */

      meth = SSLv23_client_method();
      apps_startup();
      SSLeay_add_ssl_algorithms();
      if (! bio_err)
        bio_err = BIO_new_fp(stderr,BIO_NOCLOSE);
      
      ctx = SSL_CTX_new(meth);
      if (!ctx){
          PERR("Client manager could not open an SSL context\n");
          return -1;
      }
      SSL_CTX_set_options(ctx,0);

      if (cipher)
          SSL_CTX_set_cipher_list(ctx,cipher);
  
      SSL_CTX_set_verify(ctx,verify,verify_callback);
      if (!set_cert_stuff(ctx,cert_file,key_file)) {
            SSL_CTX_free(ctx);
            return -2;
      }
  
      if (!SSL_CTX_load_verify_locations(ctx,CAfile,CApath) ||
          !SSL_CTX_set_default_verify_paths(ctx)) {
  
            PERR("Client manager no se pudo verificar el certificado\n");
            SSL_CTX_free(ctx);
            return -3;
      }
      MSG("Client manager completes SSL initializations\n");
      
      return 0;
}
#endif

int 
main (int argc, char *argv[])
{
    int err = 0;	                // returned error from routines
    int i = 0;
    struct nbd_client * self = & manager;

    if (err = cmdline (self, argc, argv), err) {
	MSG("client cmdline returned error\n");
        usage(err);
	return err;
    }

    logging ();

    /* PTB - mod_inc count will be = 1, but that's correct (dup fd) */

    self->dev = open (devicename[0], O_RDWR);
    if (self->dev < 0) {
	  PERR ("Cannot open NBD device %s: %m\n",devicename[0]);
	  exit(1);
    }
    MSG("client manager opened NBD device %s\n",devicename[0]);

#ifdef USING_SSL
    if (using_ssl) {
      err = init_ssl(self);
      if (err < 0)
        exit(1 - err);
    }
#endif

    MSG("manager detaches from shell\n");
    if (fork())   /* PTB detach */
        return 0;

    /* manager */

    setsid();
                // PTB register parent daemon
    self->pid = getpid();
    ppid = self->pid;
    setpgrp();  // PTB - this can appear to prevent connect via a make race
    self->pgid = getpgid(0);
    
    MSG("client manager started new process group %d\n", self->pgid);

                // PTB begin main loop launching children
    for (i = 0; i < nport; i++) {

        int pid = 0;

               // PTB pass on the global data from main

        init(& clients[i], i, self->dev, self->sig, self->pid, self->pgid);
        
        pid = launch(& clients[i],0);

        if (pid > 0)
          pids[i] = pid;

    }          // PTB end main loop launching children

    self->slot = -1;               // PTB to stop atexit stomping on us

    // PTB main ancestor guards children at end of launches

    microsleep(5000000);

    setsighandlers(managersighandler);  // PTB handle CHLD USR1 HUP TERM

    MSG("manager set signal handlers\n"); 

    while (1) {
        // MSG ("manager daemon sleeps\n");
        microsleep(30000000);
        if (self->dev < 0) 
          break;
        // MSG ("manager daemon wakes and syncs\n");
        ioctl (self->dev, MY_NBD_SYNC, -1);
    }

    return 0;

}

